@@ -28,12 +28,13 @@ import qualified Data.Text as T
 import qualified Data.Text.Encoding as TE
 import qualified Data.Vector as V
 import GHC.Generics (Generic)
-import System.FilePath (replaceFileName, takeBaseName, takeDirectory, (</>))
+import System.FilePath (replaceFileName, takeBaseName,
+                        takeFileName, takeDirectory, (</>))
 import Text.Parsec
 import Typst.Bind (destructuringBind)
+import Typst.Constructors (getConstructor)
 import Typst.Methods (getMethod)
-import Typst.Module.Standard (loadFileText, standardModule, symModule,
-                              sysModule)
+import Typst.Module.Standard (loadFileText, standardModule, symModule, getPath)
 import Typst.Module.Math (mathModule)
 import Typst.MathClass (mathClassOf, MathClass(Relation))
 import Typst.Parse (parseTypst)
@@ -43,8 +44,7 @@ import Typst.Syntax
 import Typst.Types
 import Typst.Util (makeFunction, nthArg)
 import qualified Toml as Toml
-import qualified Toml.FromValue as Toml
-import qualified Toml.FromValue.Generic as Toml
+import qualified Toml.Schema as Toml
 
 -- import Debug.Trace
 
@@ -59,19 +59,25 @@ evaluateTypst ::
   -- | Markup produced by 'parseTypst'
   [Markup] ->
   m (Either ParseError Content)
-evaluateTypst operations =
+evaluateTypst operations fp =
   runParserT
     (do contents <- mconcat <$> many pContent <* eof
         -- "All documents are automatically wrapped in a document element."
         pure $ Elt "document" Nothing [("body", VContent contents)])
     initialEvalState { evalOperations = operations,
-                       evalPackageRoot = "." }
+                       evalLocalDir = takeDirectory fp }
+    fp
 
 initialEvalState :: EvalState m
 initialEvalState =
-  emptyEvalState { evalIdentifiers = [(BlockScope, standardModule')] }
+  emptyEvalState { evalIdentifiers = [(BlockScope, mempty)]
+                 , evalMathIdentifiers = [(BlockScope, mathModule <> symModule)]
+                 , evalStandardIdentifiers = [(BlockScope, standardModule'')]
+                 , evalPackageRoot = "."
+                 }
   where
     standardModule' = M.insert "eval" evalFunction standardModule
+    standardModule'' = M.insert "std" (VModule "std" standardModule') standardModule'
     evalFunction = makeFunction $ do
       code :: Text <- nthArg 1
       case parseTypst "eval" ("#{\n" <> code <> "\n}") of
@@ -95,6 +101,10 @@ satisfyTok f = tokenPrim show showPos match'
 pContent :: Monad m => MP m (Seq Content)
 pContent = (pTxt <|> pElt) >>= applyShowRules >>= addTextElement
 
+applyShowRulesToVal :: Monad m => Val -> MP m Val
+applyShowRulesToVal (VContent cs) = VContent <$> applyShowRules cs
+applyShowRulesToVal x = pure x
+
 addTextElement :: Monad m => Seq Content -> MP m (Seq Content)
 addTextElement = foldM go mempty
   where
@@ -241,13 +251,9 @@ pElt = do
                 [("level", VInteger (fromIntegral level))]
           }
     Equation display ms -> inBlock BlockScope $ do
-      importModule mathModule
-      importModule symModule
-      importModule sysModule
       oldMath <- evalMath <$> getState
       updateState $ \st -> st {evalMath = True}
       content <- pInnerContents ms
-      updateState $ \st -> st {evalMath = oldMath}
       element
         "equation"
         Arguments
@@ -257,7 +263,7 @@ pElt = do
                 [ ("block", VBoolean display),
                   ("numbering", VNone)
                 ]
-          }
+          } <* updateState (\st -> st {evalMath = oldMath})
     MFrac numexp denexp -> do
       let handleParens (MGroup (Just "(") (Just ")") xs) = MGroup Nothing Nothing xs
           handleParens x = x
@@ -426,7 +432,7 @@ pExpr :: Monad m => Expr -> MP m (Seq Content)
 pExpr expr = valToContent <$> evalExpr expr
 
 evalExpr :: Monad m => Expr -> MP m Val
-evalExpr expr =
+evalExpr expr = applyShowRulesToVal =<<
   case expr of
     Literal lit -> pure $ evalLiteral lit
     Group e -> evalExpr e
@@ -569,6 +575,10 @@ evalExpr expr =
           arguments <- toArguments args
           applyElementFunction i (Function f) arguments
         VFunction Nothing _ (Function f) -> toArguments args >>= f
+        VType ty ->
+          case getConstructor ty of
+            Just (VFunction _ _ (Function f)) -> toArguments args >>= f
+            _ -> fail $ "No constructor defined for type " <> show ty
         VSymbol (Symbol _ True _) | mathMode ->
           do
             val' <- lookupIdentifier "accent"
@@ -680,6 +690,8 @@ evalExpr expr =
       case (v1, v2) of
         (VAlignment x1 y1, VAlignment x2 y2) ->
           pure $ VAlignment (x1 `mplus` x2) (y1 `mplus` y2)
+        (VArguments xs, VArguments ys) ->
+          pure $ VArguments (xs <> ys)
         _ -> case maybePlus v1 v2 of
           Nothing -> fail $ "Can't + " <> show v1 <> " and " <> show v2
           Just v -> pure v
@@ -737,6 +749,7 @@ evalExpr expr =
                 st
                   { evalShowRules =
                       ShowRule
+                        (evalNextShowRuleIdentifier st)
                         selector
                         ( \c ->
                             valToContent
@@ -746,12 +759,14 @@ evalExpr expr =
                                     named = OM.empty
                                   }
                         )
-                        : evalShowRules st
+                        : evalShowRules st,
+                    evalNextShowRuleIdentifier = evalNextShowRuleIdentifier st + 1
                   }
             _ -> updateState $ \st ->
               st
                 { evalShowRules =
                     ShowRule
+                      (evalNextShowRuleIdentifier st)
                       selector
                       ( \c ->
                           case e of
@@ -759,7 +774,8 @@ evalExpr expr =
                             Set _ _ -> pure $ Seq.singleton c
                             _ -> pure (valToContent renderVal)
                       )
-                      : evalShowRules st
+                      : evalShowRules st,
+                    evalNextShowRuleIdentifier = evalNextShowRuleIdentifier st + 1
                 }
           pure VNone
     Binding _ -> fail $ "Encountered binding out of proper context"
@@ -818,6 +834,8 @@ evalExpr expr =
         _ -> fail $ "For expression requires an Array or Dictionary"
       updateState $ \st -> st {evalFlowDirective = FlowNormal}
       go items VNone
+    Context e -> do
+      evalExpr e -- TODO for now we just ignore "context"
     Return mbe -> do
       -- these flow directives are examined in CodeBlock
       updateState (\st -> st {evalFlowDirective = FlowReturn (isJust mbe)})
@@ -930,11 +948,11 @@ data PackageConfig = PackageConfig {
 
 -- | Derived generically from record field names
 instance Toml.FromValue Config where
-  fromValue = Toml.parseTableFromValue Toml.genericParseTable
+  fromValue = Toml.genericFromTable
 
 -- | Derived generically from record field names
 instance Toml.FromValue PackageConfig where
-  fromValue = Toml.parseTableFromValue Toml.genericParseTable
+  fromValue = Toml.genericFromTable
 
 findPackageEntryPoint :: Monad m => Text -> MP m FilePath
 findPackageEntryPoint modname = do
@@ -978,7 +996,7 @@ findPackageEntryPoint modname = do
                     (localDir </> subpath) ++ "\n" ++ (cacheDir </> subpath) ++
                     "\nCompile with typst compile to bring the package into your local cache."
              -- TODO? fetch from CDN if not present in cache?
-  tomlString <- T.unpack . TE.decodeUtf8 <$> lift (loadBytes operations tomlPath)
+  tomlString <- TE.decodeUtf8 <$> lift (loadBytes operations tomlPath)
   case Toml.decode tomlString of
     Toml.Failure e -> fail (unlines ("Failure loading typst.toml" : e))
     Toml.Success _warnings cfg -> -- ignores warnings like unused keys in TOML
@@ -987,30 +1005,37 @@ findPackageEntryPoint modname = do
 loadModule :: Monad m => Text
            -> MP m (Seq Content, (Identifier, M.Map Identifier Val))
 loadModule modname = do
-  pos <- getPosition
-  (fp, modid, mbPackageRoot) <-
+  (fp, modid, mbPackageRoot, txt) <-
         if T.take 1 modname == "@"
            then do
             fp' <- findPackageEntryPoint modname
-            pure (fp',
+            operations <- evalOperations <$> getState
+            txt <- lift $ TE.decodeUtf8 <$> loadBytes operations fp'
+            pure ( takeFileName fp',
                    Identifier
                     (T.pack $ takeWhile (/= ':') . takeBaseName $
                        T.unpack modname),
-                   Just (takeDirectory fp'))
-           else if T.take 1 modname == "/" -- refers to path relative to package root
-                then do
-                  packageRoot <- evalPackageRoot <$> getState
-                  pure (packageRoot </> drop 1 (T.unpack modname),
-                        Identifier (T.pack $ takeBaseName $ T.unpack modname),
-                        Nothing)
-                else pure (replaceFileName (sourceName pos) (T.unpack modname),
-                        Identifier (T.pack $ takeBaseName $ T.unpack modname),
-                        Nothing)
-  txt <- loadFileText fp
-  case parseTypst fp txt of
+                   Just (takeDirectory fp'),
+                   txt )
+           else do
+             let fp = T.unpack modname
+             txt <- loadFileText fp
+             pure ( fp,
+                    Identifier (T.pack $ takeBaseName fp),
+                    Nothing,
+                    txt )
+  fp' <- getPath fp  -- get full path for element positions
+  case parseTypst fp' txt of
     Left err -> fail $ show err
     Right ms -> do
       operations <- evalOperations <$> getState
+      currentLocalDir <- evalLocalDir <$> getState
+      currentPackageRoot <- evalPackageRoot <$> getState
+      let (pkgroot , localdir) =
+            case mbPackageRoot of
+              Just r -> (r , takeDirectory fp)
+              Nothing -> (currentPackageRoot,
+                          currentLocalDir </> takeDirectory fp)
       res <-
         lift $
           runParserT
@@ -1021,8 +1046,8 @@ loadModule modname = do
                 pure (cs, s)
             )
             initialEvalState{evalOperations = operations,
-                             evalPackageRoot = fromMaybe (evalPackageRoot initialEvalState)
-                                                   mbPackageRoot }
+                             evalLocalDir = localdir,
+                             evalPackageRoot = pkgroot }
             fp
             ms
       case res of
@@ -1101,9 +1126,7 @@ addIdentifier ident val = do
   case identifiers of
     [] -> fail "Empty evalIdentifiers"
     ((s, i) : is) -> updateState $ \st ->
-      st
-        { evalIdentifiers = (s, M.insert ident val i) : is
-        }
+      st { evalIdentifiers = (s, M.insert ident val i) : is }
 
 updateIdentifier :: Monad m => Identifier -> Val -> MP m ()
 updateIdentifier ident val = do
