@@ -74,6 +74,11 @@ mathModule =
       makeElement (Just "math") "op" [("text", One TAny)],
       makeElement (Just "math") "underline" [("body", One TContent)],
       makeElement (Just "math") "overline" [("body", One TContent)],
+      makeElement (Just "math") "underparen" [("body", One TContent)],
+      makeElement (Just "math") "undershell" [("body", One TContent)],
+      makeElement (Just "math") "overparen" [("body", One TContent)],
+      makeElement (Just "math") "overshell" [("body", One TContent)],
+      makeElement (Just "math") "stretch" [("body", One TContent)],
       makeElement
         (Just "math")
         "underbrace"
@@ -119,7 +124,7 @@ matrixElement = VFunction (Just "mat") mempty $ Function $ \args -> do
           then as
           else as ++ [VArray (V.fromList bs)]
   -- then any leftovers
-  let fields = M.fromList [("rows", VArray (V.fromList rows))]
+  fields <- M.insert "rows" (VArray (V.fromList rows)) <$> argsToFields [] args
   pure $ VContent . Seq.singleton $ Elt "math.mat" (Just pos) fields
 
 spaceConstants :: M.Map Identifier Content
@@ -183,6 +188,7 @@ predefinedOperators =
           "hom",
           "mod",
           "ker",
+          "lcm",
           "lg",
           "ln",
           "log",
