@@ -9,14 +9,16 @@ module Typst.Module.Standard
     symModule,
     sysModule,
     loadFileText,
+    getPath,
     applyPureFunction
   )
 where
 
-import Data.Char (ord, chr)
+import Paths_typst (version)
+import Data.Version (showVersion)
 import Control.Applicative ((<|>))
 import Control.Monad (mplus, unless)
-import Control.Monad.Reader (lift, asks)
+import Control.Monad.Reader (lift, ReaderT)
 import qualified Data.Aeson as Aeson
 import qualified Data.ByteString.Lazy as BL
 import qualified Data.Csv as Csv
@@ -25,7 +27,6 @@ import qualified Data.Map.Ordered as OM
 import Data.Maybe (mapMaybe)
 import Data.Ratio ((%))
 import qualified Data.Sequence as Seq
-import qualified Data.Set as Set
 import Data.Text (Text)
 import qualified Data.Text as T
 import qualified Data.Text.Encoding as TE
@@ -38,12 +39,10 @@ import qualified Toml
 import Typst.Emoji (typstEmojis)
 import Typst.Module.Calc (calcModule)
 import Typst.Module.Math (mathModule)
-import Typst.Regex (makeRE)
 import Typst.Symbols (typstSymbols)
 import Typst.Types
 import Typst.Util
-import System.FilePath ((</>))
-import Data.List (genericTake)
+import System.FilePath.Posix ((</>), normalise)
 import Data.Time (UTCTime(..))
 import Data.Time.Calendar (fromGregorianValid)
 import Data.Time.Clock (secondsToDiffTime)
@@ -57,6 +56,7 @@ standardModule =
       ("emoji", VModule "emoji" emojiModule),
       ("calc", VModule "calc" calcModule)
     ]
+      ++ types
       ++ colors
       ++ directions
       ++ alignments
@@ -71,7 +71,11 @@ standardModule =
 
 sysModule :: M.Map Identifier Val
 sysModule =
-  M.fromList [ makeElement (Just "sys") "version" [] ]
+  M.fromList [ ("version", VVersion [0,12,0])
+             , ("inputs", VDict (OM.fromList
+                                  [("typst-hs-version",
+                                    VString (T.pack (showVersion version)))]))
+             ]
 
 symModule :: M.Map Identifier Val
 symModule = M.map VSymbol $ makeSymbolMap typstSymbols
@@ -131,11 +135,11 @@ layout =
       [ ("alignment", One TAlignment),
         ("body", One TContent)
       ],
+    makeElement Nothing "skew" [("body", One TContent)],
     makeElement Nothing "block" [("body", One TContent)],
     makeElement Nothing "box" [("body", One TContent)],
     makeElement Nothing "colbreak" [],
     makeElement Nothing "columns" [("count", One TInteger), ("body", One TContent)],
-    makeElement Nothing "grid" [("children", Many TContent)],
     makeElement Nothing "h" [("amount", One (TLength :|: TRatio :|: TFraction))],
     makeElement Nothing "v" [("amount", One (TLength :|: TRatio :|: TFraction))],
     makeElement Nothing "hide" [("body", One TContent)],
@@ -172,7 +176,24 @@ layout =
       Nothing
       "stack"
       [("children", Many (TLength :|: TRatio :|: TFraction :|: TContent))],
-    makeElement Nothing "table" [("children", Many TContent)],
+    makeElementWithScope Nothing
+      "table"
+      [("children", Many TContent)]
+      [ makeElement (Just "table") "cell" [ ("body", One TContent) ]
+      , makeElement (Just "table") "hline" []
+      , makeElement (Just "table") "vline" []
+      , makeElement (Just "table") "header" [ ("children", Many TContent) ]
+      , makeElement (Just "table") "footer" [ ("children", Many TContent) ]
+      ],
+    makeElementWithScope Nothing
+      "grid"
+      [("children", Many TContent)]
+      [ makeElement (Just "grid") "cell" [ ("body", One TContent) ]
+      , makeElement (Just "grid") "hline" []
+      , makeElement (Just "grid") "vline" []
+      , makeElement (Just "grid") "header" [ ("children", Many TContent) ]
+      , makeElement (Just "grid") "footer" [ ("children", Many TContent) ]
+      ],
     makeElementWithScope
       Nothing
       "terms"
@@ -202,7 +223,7 @@ visualize :: [(Identifier, Val)]
 visualize =
   [ makeElement Nothing "circle" [("body", One (TContent :|: TNone))],
     makeElement Nothing "ellipse" [("body", One (TContent :|: TNone))],
-    makeElement Nothing "image" [("path", One TString)],
+    makeElement Nothing "image" [("source", One (TString :|: TBytes))],
     makeElement Nothing "line" [],
     makeElement Nothing "path" [("vertices", Many TArray)],
     makeElement Nothing "polygon" [("vertices", Many TArray)],
@@ -212,10 +233,14 @@ visualize =
 
 meta :: [(Identifier, Val)]
 meta =
-  [ makeElement Nothing "bibliography" [("path", One (TString :|: TArray))],
+  [ makeElement Nothing "bibliography" [("source", One (TString :|: TArray :|: TBytes))],
     makeElement Nothing "cite" [("key", One TLabel)],
     makeElement Nothing "document" [],
-    makeElement Nothing "figure" [("body", One TContent)],
+    makeElementWithScope
+      Nothing
+      "figure"
+      [("body", One TContent)]
+      [makeElement (Just "figure") "caption" [("body", One TContent)]],
     makeElement Nothing "heading" [("body", One TContent)],
     makeElement Nothing "quote" [("body", One TContent)],
     makeElement Nothing "layout" [("func", One TFunction)],
@@ -261,6 +286,25 @@ meta =
           Failure e -> fail e)
   ]
 
+types :: [(Identifier, Val)]
+types =
+  [ ("array", VType TArray)
+  , ("bool", VType TBoolean)
+  , ("content", VType TContent)
+  , ("dictionary", VType TDict)
+  , ("int", VType TInteger)
+  , ("float", VType TFloat)
+  , ("regex", VType TRegex)
+  , ("length", VType TLength)
+  , ("alignment", VType TAlignment)
+  , ("color", VType TColor)
+  , ("symbol", VType TSymbol)
+  , ("str", VType TString)
+  , ("label", VType TLabel)
+  , ("version", VType TVersion)
+  , ("bytes", VType TBytes)
+  ]
+
 colors :: [(Identifier, Val)]
 colors =
   [ ("red", VColor $ RGB (0xff % 0xff) (0x41 % 0xff) (0x36 % 0xff) 1),
@@ -346,9 +390,6 @@ construct =
       makeFunction $
         VColor <$> (CMYK <$> nthArg 1 <*> nthArg 2 <*> nthArg 3 <*> nthArg 4)
     ),
-    ("float", makeFunction $ VFloat <$> nthArg 1),
-    ("int", makeFunction $ VInteger <$> nthArg 1),
-    ("label", makeFunction $ VLabel <$> nthArg 1),
     ( "counter",
       makeFunction $ do
         (counter :: Counter) <- nthArg 1
@@ -379,7 +420,6 @@ construct =
                           else end + 1
                       )
     ),
-    ("regex", makeFunction $ VRegex <$> (nthArg 1 >>= makeRE)),
     ( "rgb",
       makeFunction $
         VColor
@@ -392,56 +432,6 @@ construct =
                   <|> (nthArg 1 >>= hexToRGB)
               )
     ),
-    ( "str",
-      makeFunctionWithScope
-      (do
-        val <- nthArg 1
-        base <- namedArg "base" (10 :: Integer)
-        let digitVector :: V.Vector Char
-            digitVector = V.fromList $ ['0'..'9'] ++ ['A'..'Z']
-        let renderDigit n = digitVector V.!? (fromIntegral n)
-        VString <$>
-          case val of
-            VInteger n | base /= 10
-              -> case mDigits base n of
-                   Nothing -> fail "Could not convert number to base"
-                   Just ds -> maybe
-                     (fail "Could not convert number to base")
-                     (pure . T.pack)
-                     (mapM renderDigit ds)
-            _ -> fromVal val `mplus` pure (repr val))
-      [ ( "to-unicode",
-           makeFunction $ do
-             (val :: Text) <- nthArg 1
-             case T.uncons val of
-               Just (c, t) | T.null t ->
-                 pure $ VInteger $ fromIntegral $ ord c
-               _ -> fail "to-unicode expects a single character" )
-      , ( "from-unicode",
-           makeFunction $ do
-             (val :: Int) <- nthArg 1
-             pure $ VString $ T.pack [chr val] )
-      ]
-    ),
-    ( "version",
-        makeFunction $ do
-          xs <- asks positional >>= mapM fromVal
-          pure $ VVersion xs
-    ),
-    ( "symbol",
-      makeFunction $ do
-        (t :: Text) <- nthArg 1
-        vs <- drop 1 <$> allArgs
-        variants <-
-          mapM
-            ( \case
-                VArray [VString k, VString v] ->
-                  pure (Set.fromList (T.split (== '.') k), v)
-                _ -> fail "wrong type in symbol arguments"
-            )
-            vs
-        pure $ VSymbol $ Symbol t False variants
-    ),
     ( "lorem",
       makeFunction $ do
         (num :: Int) <- nthArg 1
@@ -486,17 +476,27 @@ hexToRGB _ = fail "expected string"
 loadFileLazyBytes :: Monad m => FilePath -> MP m BL.ByteString
 loadFileLazyBytes fp = do
   operations <- evalOperations <$> getState
-  root <- evalPackageRoot <$> getState
-  lift $ BL.fromStrict <$> loadBytes operations (root </> fp)
+  path <- getPath fp
+  lift $ BL.fromStrict <$> loadBytes operations path
 
 loadFileText :: Monad m => FilePath -> MP m T.Text
 loadFileText fp = do
   operations <- evalOperations <$> getState
+  path <- getPath fp
+  lift $ TE.decodeUtf8 <$> loadBytes operations path
+
+-- a leading / = relative to package root
+getPath :: Monad m => FilePath -> MP m FilePath
+getPath ('/':fp') = do
   root <- evalPackageRoot <$> getState
-  lift $ TE.decodeUtf8 <$> loadBytes operations (root </> fp)
+  pure $ root </> fp'
+getPath fp = do
+  pkgroot <- evalPackageRoot <$> getState
+  localdir <- evalLocalDir <$> getState
+  pure $ normalise $ pkgroot </> localdir </> fp
 
 getUTCTime :: Monad m => MP m UTCTime
-getUTCTime = (currentUTCTime . evalOperations <$> getState) >>= lift
+getUTCTime = getState >>= lift . currentUTCTime . evalOperations
 
 time :: [(Identifier, Val)]
 time =
@@ -535,16 +535,14 @@ dataLoading =
     ),
     ( "json",
       makeFunction $ do
-        fp <- nthArg 1
-        bs <- lift $ loadFileLazyBytes fp
+        bs <- getFileOrBytes
         case Aeson.eitherDecode bs of
           Left e -> fail e
           Right (v :: Val) -> pure v
     ),
     ( "yaml",
       makeFunction $ do
-        fp <- nthArg 1
-        bs <- lift $ loadFileLazyBytes fp
+        bs <- getFileOrBytes
         case Yaml.decodeEither' (BL.toStrict bs) of
           Left e -> fail $ show e
           Right (v :: Val) -> pure v
@@ -552,21 +550,23 @@ dataLoading =
     ( "read",
       makeFunction $ do
         fp <- nthArg 1
-        t <- lift $ loadFileText fp
-        pure $ VString t
+        enc <- namedArg "encoding" (VString "utf-8")
+        case enc of
+          VNone -> do bs <- lift $ loadFileLazyBytes fp
+                      pure $ VBytes $ BL.toStrict bs
+          _ -> do t <- lift $ loadFileText fp
+                  pure $ VString t
     ),
     ( "toml",
       makeFunction $ do
-        fp <- nthArg 1
-        t <- lift $ loadFileText fp
-        case Toml.decode (T.unpack t) of
+        bs <- getFileOrBytes
+        case Toml.decode (TE.decodeUtf8 $ BL.toStrict bs) of
           Toml.Failure e -> fail (unlines ("toml errors:" : e))
           Toml.Success _ v -> pure v
     ),
     ( "xml",
       makeFunction $ do
-        fp <- nthArg 1
-        bs <- lift $ loadFileLazyBytes fp
+        bs <- getFileOrBytes
         case XML.parseLBS XML.def bs of
           Left e -> fail $ show e
           Right doc ->
@@ -611,29 +611,15 @@ applyPureFunction (Function f) vals =
 
 initialEvalState :: MonadFail m => EvalState m
 initialEvalState =
-  emptyEvalState { evalIdentifiers = [(BlockScope, standardModule)] }
+  emptyEvalState { evalIdentifiers = [(BlockScope, mempty)]
+                 , evalMathIdentifiers = [(BlockScope, mathModule <> symModule)]
+                 , evalStandardIdentifiers = [(BlockScope, standardModule)]
+                 }
 
--- mDigitsRev, mDigits from the unmaintained digits package
--- https://hackage.haskell.org/package/digits-0.3.1
--- (c) 2009-2016 Henry Bucklow, Charlie Harvey -- BSD-3-Clause license.
-mDigitsRev :: Integral n
-    => n         -- ^ The base to use.
-    -> n         -- ^ The number to convert to digit form.
-    -> Maybe [n] -- ^ Nothing or Just the digits of the number in list form, in reverse.
-mDigitsRev base i = if base < 1
-                    then Nothing -- We do not support zero or negative bases
-                    else Just $ dr base i
-    where
-      dr _ 0 = []
-      dr b x = case base of
-                1 -> genericTake x $ repeat 1
-                _ -> let (rest, lastDigit) = quotRem x b
-                     in lastDigit : dr b rest
-
--- | Returns the digits of a positive integer as a Maybe list.
---   or Nothing if a zero or negative base is given
-mDigits :: Integral n
-    => n -- ^ The base to use.
-    -> n -- ^ The number to convert to digit form.
-    -> Maybe [n] -- ^ Nothing or Just the digits of the number in list form
-mDigits base i = reverse <$> mDigitsRev base i
+getFileOrBytes :: Monad m => ReaderT Arguments (MP m) BL.ByteString
+getFileOrBytes = do
+  v <- nthArg 1
+  case v of
+    VString fp -> lift $ loadFileLazyBytes (T.unpack fp)
+    VBytes bs -> pure $ BL.fromStrict bs
+    _ -> fail "expecting file path or bytes"
