@@ -68,11 +68,12 @@ import qualified Data.Set as Set
 import Data.String
 import Data.Text (Text)
 import qualified Data.Text as T
+import Data.ByteString (ByteString)
 import Data.Vector (Vector)
 import qualified Data.Vector as V
 import Text.Parsec
 import qualified Toml
-import qualified Toml.FromValue as Toml
+import qualified Toml.Schema as Toml
 import qualified Toml.Pretty as Toml
 import qualified Text.PrettyPrint as P
 import Text.Read (readMaybe)
@@ -82,37 +83,69 @@ import Data.Time (UTCTime, Day, DiffTime, timeOfDayToTime, localDay, localTimeOf
 import Data.Time.Format (defaultTimeLocale, formatTime)
 import System.Directory (XdgDirectory(..))
 
+-- | A Typst value. More documentation can be found in the 
+-- [Foundations chapter](https://typst.app/docs/reference/foundations/)
+-- of the Typst reference manual. A more concise (but somewhat outdated) 
+-- summary can also be found in 
+-- [L. Mädje "Typst: a programmable markup language for typesetting", page 32-33](https://www.user.tu-berlin.de/laurmaedje/programmable-markup-language-for-typesetting.pdf).
 data Val
+  -- | The @none@ value, indicates the absence of any other value.
   = VNone
+  -- | The @auto@ value, used to automatically set an appropriate value.
   | VAuto
+  -- | A @bool@ value.
   | VBoolean !Bool
+  -- | An @int@ value.
   | VInteger !Integer
+  -- | A @float@ value.
   | VFloat !Double
+  -- | A @ratio@ value, a proportion of a certain whole, for example @50%@.
   | VRatio !Rational
+  -- | A @length@ or a @relative@ value.
   | VLength !Length
+  -- | An @alignment@ value, indicating the alignment of some content along both
+  -- the horizontal and vertical axes.
   | VAlignment (Maybe Horiz) (Maybe Vert)
-  | VAngle !Double -- degrees
+  -- | An @angle@ value (expressed internally in degrees).
+  | VAngle !Double
+  -- | A @fraction@ value, defining the proportions of remaing space is 
+  -- to be distributed, e.g. @2 fr@.
   | VFraction !Double
+  -- | A @color@ value. Not all Typst color spaces are supported; 
+  -- only @rgb@, @cmyk@, and @luma@ are available. 
+  -- See issue [#35](https://github.com/jgm/typst-hs/issues/35#issuecomment-1926182040).
   | VColor !Color
+  -- | A @symbol@ value, representing a Unicode symbol.
   | VSymbol !Symbol
+  -- | A UTF-8 encoded text @string@.
   | VString !Text
+  -- | A @regex@ (regular expression). See 'RE' for details.
   | VRegex !RE
+  -- | A @datetime@ value, a date, a time, or a combination of both.
   | VDateTime (Maybe Day) (Maybe DiffTime)
+  -- | A @content@ value, see 'Content' for more details.
   | VContent (Seq Content)
+  -- | An @array@ value, for example @(10, 20, 30)@.
   | VArray (Vector Val)
+  -- | A @dictionary@ value, for example @(a:20, b:30)@.
   | VDict (OM.OMap Identifier Val)
   | VTermItem (Seq Content) (Seq Content)
+  -- | A @direction@ to lay out content.
   | VDirection Direction
+  -- | A Typst function.
   | VFunction (Maybe Identifier) (M.Map Identifier Val) Function
   | -- first param is Just ident if element function
     -- second param is a map of subfunctions in this function's scope
+    -- | Positional and named function arguments
     VArguments Arguments
+  -- | A @label@ to some element, for example @<hello>@. 
   | VLabel !Text
   | VCounter !Counter
   | VSelector !Selector
   | VModule Identifier (M.Map Identifier Val)
   | VStyles -- just a placeholder for now
   | VVersion [Integer]
+  | VBytes ByteString
   | VType !ValType
   deriving (Show, Eq, Typeable)
 
@@ -129,19 +162,20 @@ instance FromJSON Val where
 instance Toml.FromValue Val where
   fromValue = pure . tomlToVal
 
-tomlToVal :: Toml.Value -> Val
-tomlToVal (Toml.Bool x) = VBoolean x
-tomlToVal (Toml.Integer x) = VInteger x
-tomlToVal (Toml.String x) = VString (T.pack x)
-tomlToVal (Toml.Float x) = VFloat x
-tomlToVal (Toml.TimeOfDay x) = VDateTime Nothing (Just (timeOfDayToTime x))
-tomlToVal (Toml.Day x) = VDateTime (Just x) Nothing
-tomlToVal (Toml.LocalTime x) = VDateTime (Just (localDay x)) (Just (timeOfDayToTime (localTimeOfDay x)))
-tomlToVal (Toml.Array x) = VArray (V.fromList (map tomlToVal x))
-tomlToVal (Toml.Table x) = VDict (OM.fromList [(Identifier (T.pack k), tomlToVal v) | (k,v) <- M.assocs x])
+tomlToVal :: Toml.Value' a -> Val
+tomlToVal (Toml.Bool' _ x) = VBoolean x
+tomlToVal (Toml.Integer' _ x) = VInteger x
+tomlToVal (Toml.Text' _ x) = VString x
+tomlToVal (Toml.Double' _ x) = VFloat x
+tomlToVal (Toml.TimeOfDay' _ x) = VDateTime Nothing (Just (timeOfDayToTime x))
+tomlToVal (Toml.Day' _ x) = VDateTime (Just x) Nothing
+tomlToVal (Toml.LocalTime' _ x) = VDateTime (Just (localDay x)) (Just (timeOfDayToTime (localTimeOfDay x)))
+tomlToVal (Toml.List' _ x) = VArray (V.fromList (map tomlToVal x))
+tomlToVal (Toml.Table' _ (Toml.MkTable x)) = VDict (OM.fromList [(Identifier k, tomlToVal v) | (k,(_,v)) <- M.assocs x])
   -- typst specifies that unsupported datetimes map to strings and we don't have a place for the timezone
-tomlToVal v@Toml.ZonedTime{} = VString (T.pack (show (Toml.prettyValue v)))
+tomlToVal v@Toml.ZonedTime'{} = VString (T.pack (show (Toml.prettyValue v)))
 
+-- | A Typst type, see documentation for 'Val'.
 data ValType
   = TNone
   | TAuto
@@ -172,6 +206,7 @@ data ValType
   | TCounter
   | TLocation
   | TVersion
+  | TBytes
   | TType
   | TAny
   | ValType :|: ValType
@@ -208,6 +243,7 @@ valType v =
     VSelector {} -> TSelector
     VStyles {} -> TStyles
     VVersion {} -> TVersion
+    VBytes {} -> TBytes
     VType {} -> TType
 
 hasType :: ValType -> Val -> Bool
@@ -574,24 +610,36 @@ data Operations m =
 data EvalState m = EvalState
   { evalIdentifiers :: [(Scope, M.Map Identifier Val)],
     -- first item is current block, then superordinate block, etc.
+    -- The standard identifiers and the identifiers that
+    -- are imported by default into math contexts are special,
+    -- since both can be overridden by user-defined identifiers.
+    -- So, we store them separately.
+    evalStandardIdentifiers :: [(Scope, M.Map Identifier Val)],
+    evalMathIdentifiers :: [(Scope, M.Map Identifier Val)],
     evalCounters :: M.Map Counter Integer,
     evalMath :: Bool,
     evalShowRules :: [ShowRule],
+    evalNextShowRuleIdentifier :: Int,
     evalStyles :: M.Map Identifier Arguments,
     evalFlowDirective :: FlowDirective,
     evalPackageRoot :: FilePath,
+    evalLocalDir :: FilePath,
     evalOperations :: Operations m
   }
 
 emptyEvalState :: EvalState m
 emptyEvalState = EvalState
     { evalIdentifiers = [],
+      evalStandardIdentifiers = [],
+      evalMathIdentifiers = [],
       evalCounters = mempty,
       evalMath = False,
       evalShowRules = [],
+      evalNextShowRuleIdentifier = 1,
       evalStyles = mempty,
       evalFlowDirective = FlowNormal,
       evalPackageRoot = mempty,
+      evalLocalDir = mempty,
       evalOperations = undefined
     }
 
@@ -619,10 +667,14 @@ instance MonadFail Attempt where
   fail = Failure
 
 data ShowRule
-  = ShowRule Selector (forall m. Monad m => Content -> MP m (Seq Content))
+  = ShowRule Int Selector (forall m. Monad m => Content -> MP m (Seq Content))
 
 instance Show ShowRule where
-  show (ShowRule sel _) = "ShowRule " <> show sel <> " <function>"
+  show (ShowRule ident sel _) =
+    "ShowRule " <> show ident <> " " <> show sel <> " <function>"
+
+instance Eq ShowRule where
+  (ShowRule id1 _ _) == (ShowRule id2 _ _) = id1 == id2
 
 type MP m = ParsecT [Markup] (EvalState m) m
 
@@ -768,7 +820,11 @@ data Color
   | Luma Rational
   deriving (Show, Eq, Ord, Typeable)
 
-data Direction = Ltr | Rtl | Ttb | Btt
+data Direction 
+  = Ltr -- ^ Left to right
+  | Rtl -- ^ Right to left
+  | Ttb -- ^ Top to bottom
+  | Btt -- ^ Bottom to top
   deriving (Show, Eq, Ord, Typeable)
 
 prettyVal :: Val -> P.Doc
@@ -870,6 +926,7 @@ prettyVal expr =
     VSelector _ -> mempty
     VStyles -> mempty
     VVersion xs -> text $ T.intercalate "." (map (T.pack . show) xs)
+    VBytes bs -> text $ "bytes(" <> T.pack (show (BS.length bs)) <> ")"
     VType ty -> text $ prettyType ty
 
 prettyType :: ValType -> Text
@@ -941,4 +998,12 @@ lookupIdentifier ident = do
       go ((_, i) : is) = case M.lookup ident i of
         Just v -> pure v
         Nothing -> go is
-  getState >>= go . evalIdentifiers
+  identifiers <- evalIdentifiers <$> getState
+  mathIdentifiers <- evalMathIdentifiers <$> getState
+  standardIdentifiers <- evalStandardIdentifiers <$> getState
+  math <- evalMath <$> getState
+  go $ case identifiers of
+         -- in math mode, we insert the sym and math modules right before
+         -- the final module (standard module)
+         (_:_) | math -> identifiers ++ mathIdentifiers ++ standardIdentifiers
+         _ -> identifiers ++ standardIdentifiers
